import fs from 'fs'

const INFILE = process.env['INFILE'],
  OUTFILE = process.env['OUTFILE']

export default function scuttlebuttServer(server) {
  const primusServer = new (require('primus'))(server, {}),
      Dispatcher = require('./dispatcher').default,
      gossip = new Dispatcher()

  const statistics = {}

  var statisticsDirty = true

  // prime statistics for when spark.id is undefined, presumably server messages
  statistics[undefined] = {
    recv: 0, sent: 0, s: 'other'
  }

  setInterval(() => {
    if (!statisticsDirty)
      return

    statisticsDirty = false

    /*
    // full client statistics
    console.log('^^^ ' + (new Date()) + ' ^^^')
    for (let spark in statistics) {
      console.log(`${spark}: ${statistics[spark].recv} recv ${statistics[spark].sent} sent (${statistics[spark].s})`)
    }
    */

    // basic statistics
    console.log([
      (new Date()).toLocaleString('en-AU'),
      ': ',
      (() => {
        let recv = 0, sent = 0, connected = 0, disconnected = 0, other = 0
        for (let spark in statistics) {
          recv += statistics[spark].recv
          sent += statistics[spark].sent

          if (statistics[spark].s === 'connected')
            connected++
          else if (statistics[spark].s === 'disconnected')
            disconnected++
          else
            other++
        }

        return `recv ${recv}, sent ${sent}, (${connected} 🌏, ${disconnected} 🔕, ${other} 👥)`
      })()
    ].join(''))

  }, 10000)

  connectRedux(gossip)

  if (INFILE || OUTFILE) {
    if (INFILE) {
      const gossipWriteSteam = gossip.createWriteStream()
      fs.createReadStream(INFILE).pipe(gossipWriteSteam)

      console.log('📼  Reading from ' + INFILE)
    }

    if (OUTFILE) {
      const gossipReadSteam = gossip.createReadStream()
      let hasSynced = false

      // For some reason, we're not getting any 'sync' events from Dispatcher,
      // so we'll listen for it in the datastream and write to disk after it.
      // <https://github.com/dominictarr/scuttlebutt#persistence>

      gossipReadSteam.on('data', (data) => {
        if (data === '"SYNC"\n') {
          if (hasSynced) {
            console.log('📼  Write history (sync) 🚨🚨🚨🚨 RECEIVED MULTIPLE SYNCS')
          } else {
            console.log('📼  Writing to ' + OUTFILE)
            gossipReadSteam.pipe(fs.createWriteStream(OUTFILE))
            hasSynced = true
          }
        }
      })

      gossip.on('sync', function () {
        gossipReadSteam.pipe(fs.createWriteStream(OUTFILE))
      })

      console.log('📼  Ready to write to ' + OUTFILE)
    }
  }

  primusServer.on('connection', (spark) => {
    var stream = gossip.createStream()

    // console.log('[io] connection', spark.address, spark.id)

    statistics[spark.id] = {
      recv: 0, sent: 0, s: 'connected'
    }

    spark.on('data', function recv(data) {
      // console.log('[io]', spark.id, '<-', data);
      statistics[spark.id] ? statistics[spark.id].recv++ : console.log(`${spark.id} didn't have recv stats ready?`)
      statisticsDirty = true
      // statistics[spark.id].recv++
      stream.write(data)
    });

    stream.on('data', (data) => {
      // console.log('[io]', spark.id || 'origin', '->', data);
      statistics[spark.id] ? statistics[spark.id].sent++ : console.log(`${spark.id} didn't have sent stats ready?`)
      statisticsDirty = true
      spark.write(data)
    })

    stream.on('error', (error) => {
      console.log('[io]', spark.id, 'ERROR:', error);
      spark.end('Disconnecting due to error', { reconnect: true })
    })
  })

  primusServer.on('disconnection', (spark) => {
    statistics[spark.id].s = 'disconnected'
    // delete statistics[spark.id]
  })

}

function connectRedux(gossip) {
  const Redux = require('redux'),
    reducer = (state = [], action) => state.concat(action),
    store = Redux.createStore(gossip.wrapReducer(reducer), undefined),
    dispatch = gossip.wrapDispatch(store.dispatch)

  // other things we might want to do ->
  // store.subscribe(render)
}
